// Copyright (C) 1994-98 by Addison-Wesley Publishing Co., Inc.
//
// This code is provided for its instructional value.
// It has been tested with care, but it is not guaranteed
// for any particular purpose.  The publisher and authors
// do not offer any warranties or representations, nor do
// they accept any liabilities with respect to this code.

#include "cpp-faqs.hpp"

#include <cassert>
#include <string>
using namespace std;


class Fred {
public:
  static Fred create1(string s, int i);  // Named constructors; see FAQ 16.08
  static Fred create2(float x, float y);
  Fred(const Fred& f);
  Fred& operator= (const Fred& f);
 ~Fred();
  void sampleInspectorMethod() const;   // Does not change this object
  void sampleMutatorMethod();           // Changes this object

private:

  class Data {
  public:
    Data() : count_(1) { }
    Data(const Data& d) : count_(1) { }              // Don't copy count_!
    Data& operator= (const Data&) { return *this; }  // Don't copy count_!
    virtual ~Data() { assert(count_ == 0); }         // Virtual destructor
    virtual Data* clone() const = 0;                 // Virtual constructor
    virtual void sampleInspectorMethod() const = 0;  // See FAQ 21.11
    virtual void sampleMutatorMethod() = 0;
  private:
    unsigned count_;   // count_ doesn't need to be protected
    friend Fred;
  };

  class Der1 : public Data {
  public:
    Der1(string s, int i);
    virtual void sampleInspectorMethod() const;
    virtual void sampleMutatorMethod();
    virtual Data* clone() const;
    // ...
  };

  class Der2 : public Data {
  public:
    Der2(float x, float y);
    virtual void sampleInspectorMethod() const;
    virtual void sampleMutatorMethod();
    virtual Data* clone() const;
    // ...
  };

  Fred(Data* data);
  // Creates a Fred smart-reference that owns *data
  // It is private to force users to use a createXXX() member function
  // Requirement: data must not be NULL

  Data* data_;   // Invariant: data_ is never NULL
  friend Der1;
  friend Der2;
};

Fred::Fred(Data* data)    : data_(data)    { assert(data != NULL); }
Fred::Fred(const Fred& f) : data_(f.data_) { ++ data_->count_; }

Fred Fred::create1(string s, int i)   { return Fred(new Der1(s, i)); }
Fred Fred::create2(float x, float y)  { return Fred(new Der2(x, y)); }

Fred::Data* Fred::Der1::clone() const { return new Der1(*this); }
Fred::Data* Fred::Der2::clone() const { return new Der2(*this); }

Fred& Fred::operator= (const Fred& f)
{
  // DO NOT CHANGE THE ORDER OF THESE STATEMENTS!
  // (This order properly handles self-assignment; See FAQ 24.03)
  ++ f.data_->count_;
  if (--data_->count_ == 0) delete data_;
  data_ = f.data_;
  return *this;
}

Fred::~Fred()
{ if (--data_->count_ == 0) delete data_; }

void Fred::sampleInspectorMethod() const
{
  // This member function promises ("const") not to change anything in *data_
  // Therefore we simply "pass the member function through" to *data_:
  data_->sampleInspectorMethod();
}

void Fred::sampleMutatorMethod()
{
  // This member function might need to change things in *data_
  // Thus it first checks if this is the only pointer to *data_
  if (data_->count_ > 1) {
    Data* d = data_->clone();   // Virtual Constructor Idiom; see FAQ 21.07
    -- data_->count_;
    data_ = d;
  }
  assert(data_->count_ == 1);

  // Now we "pass the member function through" to *data_:
  data_->sampleMutatorMethod();
}
